/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sgp;

import java.util.ArrayList;
import java.util.Random;

/**
 *
 * @author Fanda
 */
public class Individual implements Cloneable{

    public double fitness = 0.0;
    public Node treeRoot;

    /**
     * Constructor.
     */
    public Individual() {
        this.initIndividual(SGP.functions, SGP.terminals);
    }
    
    /**
     * Constructor.
     */
    public Individual(boolean f) {
        
    }
    
    /**
     * Initializes individual by BFS method.
     *
     * @param f possible inner nodes
     * @param t possible terminal nodes
     */
    public void initIndividual(ArrayList<Node> f, ArrayList<Node> t) {
        int actualDepth = 0;
        ArrayList<Node> queue = new ArrayList(); // BFS queue.
        Random r = new Random();
        // Randomly choose root from functions.
        treeRoot = f.get(r.nextInt(f.size())).clone();
        queue.add(treeRoot);
        Node expanded;              // Expanded node.

        while (!queue.isEmpty()) {
            expanded = queue.get(0); // Pop from queue.
            queue.remove(0);
            actualDepth = expanded.depth + 1;
            // Create children of expanded node.
            for (int i = 0; i < expanded.arity; i++) {
                // If depth is lower than minimal, choose from functions only.
                if (actualDepth < SGP.MIN_INIT_DEPTH) {
                    expanded.children[i] = f.get(r.nextInt(f.size())).clone();
                    expanded.children[i].depth = actualDepth;
                    expanded.children[i].parent = expanded;
                    queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                } // Else if depth is greater than minimal and lower than maximal, choose from function and terminals.
                else if (actualDepth < SGP.MAX_INIT_DEPTH) {
                    int id = r.nextInt(f.size() + t.size());
                    // Determine if random int points to terminals or functions.
                    if (id >= f.size()) {
                        // Terminal can't be expanded, so do not add it to the queue.
                        expanded.children[i] = t.get(id - f.size()).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                    } else {
                        expanded.children[i] = f.get(id).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                        queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                    }
                }
                // Else maximal depth is reached, choose only from terminals.
                else {
                    expanded.children[i] = t.get(r.nextInt(t.size())).clone();
                    expanded.children[i].depth = actualDepth;
                    expanded.children[i].parent = expanded;
                }
            }
        }
    }
    
    public static void onePointCrossover(Individual ind1, Individual ind2) {
        Random r = new Random();
        int rnd1 = r.nextInt(100);
        int rnd2 = r.nextInt(100);
        int counter = 0;
        ArrayList<Node> queue1 = new ArrayList();
        ArrayList<Node> queue2 = new ArrayList();
        queue1.add(ind1.treeRoot);
        queue2.add(ind2.treeRoot);
        Node actual1 = queue1.get(0);
        Node actual2 = queue2.get(0);
        
        int child1ID = -1;
        int child2ID = -1;
        boolean f = false;
        // Go through tree by BFS and count nodes.
        while(counter < rnd1) {
            child1ID = 0;
            counter++;
            actual1 = queue1.get(0);
            queue1.remove(0);
            for(Node n : actual1.children) {
                queue1.add(n);
                child1ID++;
                if(counter + child1ID == rnd1) {
                    child1ID--;
                    actual1 = n;
                    f = true;
                    break;
                }
            }
            if(f) {
                break;
            }
            child1ID = -1;
            if(queue1.isEmpty()) {
                queue1.add(ind1.treeRoot);
            }
        }
        counter = 0;
        f = false;
        while(counter < rnd2) {
            child2ID = 0;
            counter++;
            actual2 = queue2.get(0);
            queue2.remove(0);
            for(Node n : actual2.children) {
                queue2.add(n);
                child2ID++;
                if(counter + child2ID == rnd2) {
                    child2ID--;
                    actual2 = n;
                    f = true;
                    break;
                }
            }
            if(f) {
                break;
            }
            child2ID = -1;
            if(queue2.isEmpty()) {
                queue2.add(ind2.treeRoot);
            }
        }
        
        Node parent1 = actual1.parent;
        Node parent2 = actual2.parent;
        
        // Root node.
        if(parent1 == null) {
            ind1.treeRoot = actual2.clone();
            for(int c = 0; c < ind1.treeRoot.arity; c++) {
                ind1.treeRoot.children[c] =actual2.children[c];
                actual2.children[c] = ind1.treeRoot;
            }
        }
        else {
            System.out.println(1 + " " + actual1.parent.name + " => " + actual1.name);

            // Switch parent and children of the cross point.
            Node newChild;
            parent1.children[child1ID] = actual2.clone();
            newChild = parent1.children[child1ID];
            newChild.parent = parent1;
            for(int c = 0; c < newChild.arity; c++) {
                newChild.children[c] = actual2.children[c];
                actual1.children[c].parent = newChild;
            }                   
        }

        if(parent2 == null) {
            ind2.treeRoot = actual1.clone();
            for(int c = 0; c < ind2.treeRoot.arity; c++) {
                ind2.treeRoot.children[c] = actual1.children[c];
                actual1.children[c] = ind2.treeRoot;
            }
        }
        else {
            System.out.println(2 + " " + actual2.parent.name + " => " + actual2.name);
            Node newChild;
            parent2.children[child2ID] = actual1.clone();
            newChild = parent2.children[child2ID];
            newChild.parent = parent2;
            for(int c = 0; c < newChild.arity; c++) {
                newChild.children[c] = actual1.children[c];
                actual1.children[c].parent = newChild;
            }                   
        }
        
    }
    
    @Override
    public Individual clone() {
        Individual ind = new Individual(false); // Empty individual;
        ind.treeRoot = this.treeRoot.clone();
        
        ArrayList<Node> origQueue = new ArrayList(); // BFS queue.
        ArrayList<Node> newQueue = new ArrayList(); // BFS queue.
        origQueue.add(this.treeRoot);
        newQueue.add(ind.treeRoot);
        Node origExpanded;              // Expanded node.
        Node newExpanded;
        
        while (!origQueue.isEmpty()) {
            origExpanded = origQueue.get(0); // Pop from queue.
            newExpanded = newQueue.get(0); // Pop from queue.
            origQueue.remove(0);
            newQueue.remove(0);
            // Create children of expanded node.
            for (int i = 0; i < origExpanded.arity; i++) {
                newExpanded.children[i] = origExpanded.children[i].clone();
                newExpanded.children[i].depth = newExpanded.depth+1;
                newExpanded.children[i].parent = newExpanded;
                origQueue.add(origExpanded.children[i]);
                newQueue.add(newExpanded.children[i]);
            }
        }
        return ind;
    }
    
    public String toString() {
        String r = "";
        Node n = treeRoot;
        ArrayList<Node> queue = new ArrayList();
        queue.add(n);
        Node expand;
        while(!queue.isEmpty()) {
            expand = queue.get(0);
            queue.remove(0);
            
            r += expand.name + " ";
            
            for(int i = 0; i < expand.arity; i++) {
                queue.add(expand.children[i]);
            }
        }
        
        return r;
    }
}
