/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sgp;

import java.util.ArrayList;
import java.util.Random;

/**
 *
 * @author Fanda
 */
public class Individual implements Cloneable, Comparable {

    public double fitness = 0.0;
    public Node rootNode;
    public String treeString;

    /**
     * Constructor.
     */
    public Individual() {
        this.initIndividual(SGP.functions, SGP.terminals);
    }

    /**
     * Constructor.
     */
    public Individual(boolean f) {

    }

    /**
     * Execution method. Gets results of the function for the points defined in
     * Regression.java and compares its values.
     */
    public void execute() {
        this.fitness = 0.0;
        for (int i = 0; i < Regression.results.length; i++) {
            double result = this.rootNode.executeNode(i);
            this.fitness += (Math.abs(result - Regression.getY(i)));
        }

    }

    /**
     * Initializes individual by BFS method.
     *
     * @param f possible inner nodes
     * @param t possible terminal nodes
     */
    public void initIndividual(ArrayList<Node> f, ArrayList<Node> t) {
        int actualDepth = 0;
        ArrayList<Node> queue = new ArrayList(); // BFS queue.
        Random r = new Random();
        // Randomly choose root from functions.
        rootNode = f.get(r.nextInt(f.size())).clone();
        queue.add(rootNode);
        Node expanded;              // Expanded node.

        while (!queue.isEmpty()) {
            expanded = queue.get(0); // Pop from queue.
            queue.remove(0);
            actualDepth = expanded.depth + 1;
            // Create children of expanded node.
            for (int i = 0; i < expanded.arity; i++) {
                // If depth is lower than minimal, choose from functions only.
                if (actualDepth < SGP.MIN_INIT_DEPTH) {
                    expanded.children[i] = f.get(r.nextInt(f.size())).clone();
                    expanded.children[i].depth = actualDepth;
                    expanded.children[i].parent = expanded;
                    queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                } // Else if depth is greater than minimal and lower than maximal, choose from function and terminals.
                else if (actualDepth < SGP.MAX_INIT_DEPTH) {
                    int id = r.nextInt(f.size() + t.size());
                    // Determine if random int points to terminals or functions.
                    if (id >= f.size()) {
                        // Terminal can't be expanded, so do not add it to the queue.
                        expanded.children[i] = t.get(id - f.size()).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                    } else {
                        expanded.children[i] = f.get(id).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                        queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                    }
                } // Else maximal depth is reached, choose only from terminals.
                else {
                    expanded.children[i] = t.get(r.nextInt(t.size())).clone();
                    expanded.children[i].depth = actualDepth;
                    expanded.children[i].parent = expanded;
                }
            }
        }
    }

    /**
     * One point crossover operator. Method rewrites structure of input individual!!!
     * @param ind1 first parent.
     * @param ind2 second parent
     */
    public static void onePointCrossover(Individual ind1, Individual ind2) {
        Random r = new Random();
        int rnd = r.nextInt(100);
        int counter = 0;
        ArrayList<Node> queue1 = new ArrayList();
        ArrayList<Node> queue2 = new ArrayList();
        queue1.add(ind1.rootNode);
        queue2.add(ind2.rootNode);
        Node actual1 = queue1.get(0);
        Node actual2 = queue2.get(0);

        // Go through tree by BFS and count nodes.
        while (counter < rnd) {
            counter++;
            actual1 = queue1.get(0);
            queue1.remove(0);
            actual2 = queue2.get(0);
            queue2.remove(0);
            for (Node n : actual1.children) {
                queue1.add(n);
            }
            if (queue1.isEmpty()) {
                queue1.add(ind1.rootNode);
            }
            if (queue2.isEmpty()) {
                queue2.add(ind2.rootNode);
            }
        }

        Node parent1 = actual1.parent;
        Node parent2 = actual2.parent;

        //System.out.println(actual1.name);
        //System.out.println(actual2.name);
        // Root node.
        if (parent1 == null) {
            ind1.rootNode = actual2.clone();
            ind1.rootNode.parent = null;
            for (int c = 0; c < ind1.rootNode.arity; c++) {
                ind1.rootNode.children[c] = actual2.children[c];
                actual2.children[c].parent = ind1.rootNode;
            }
        } else {
            // Swithch parent and children of the cross point.
            int childID = -1;
            for (int child = 0; child < parent1.children.length; child++) {
                if (parent1.children[child] == actual1) {
                    childID = child;
                    break;
                }
            }
            Node newChild;
            parent1.children[childID] = actual2.clone();
            newChild = parent1.children[childID];
            newChild.parent = parent1;
            for (int c = 0; c < newChild.arity; c++) {
                newChild.children[c] = actual2.children[c];
                actual2.children[c].parent = newChild;
            }
        }

        if (parent2 == null) {
            ind2.rootNode = actual1.clone();
            ind2.rootNode.parent = null;
            for (int c = 0; c < ind2.rootNode.arity; c++) {
                ind2.rootNode.children[c] = actual1.children[c];
                actual1.children[c].parent = ind2.rootNode;
            }
        } else {
            //System.out.println(1 + " " + actual2.parent.name);
            int childID = -1;
            for (int child = 0; child < parent2.children.length; child++) {
                if (parent2.children[child] == actual1) {
                    childID = child;
                    break;
                }
            }
            Node newChild;
            parent2.children[childID] = actual1.clone();
            newChild = parent2.children[childID];
            newChild.parent = parent2;
            for (int c = 0; c < newChild.arity; c++) {
                newChild.children[c] = actual1.children[c];
                actual1.children[c].parent = newChild;
            }
        }

    }

    /**
     * Subtree mutation. Method rewrites structure of input individual!!!
     * @param ind mutated individual
     */
    public static void mutation(Individual ind) {
        Random r = new Random();
        int rnd = r.nextInt(100);
        int counter = 0;
        ArrayList<Node> queue1 = new ArrayList();
        queue1.add(ind.rootNode);
        Node actual1 = queue1.get(0);

        // Go through tree by BFS and count nodes.
        while (counter < rnd) {
            counter++;
            actual1 = queue1.get(0);
            queue1.remove(0);
            for (Node n : actual1.children) {
                queue1.add(n);
            }
            if (queue1.isEmpty()) {
                queue1.add(ind.rootNode);
            }
        }

        Node parent1 = actual1.parent;

        // Root node, generate whole new individual.
        if (parent1 == null) {
            ind.initIndividual(SGP.functions, SGP.terminals);
        } else {

            int childID = -1;
            for (int child = 0; child < parent1.children.length; child++) {
                if (parent1.children[child] == actual1) {
                    childID = child;
                    break;
                }
            }

            ArrayList<Node> f = SGP.functions;
            ArrayList<Node> t = SGP.terminals;
            int id = r.nextInt(f.size() + t.size());
            Node newChild;
            if (parent1.depth + 1 < SGP.MAX_INIT_DEPTH) {
                // Terminal node.
                if (id >= f.size()) {
                    parent1.children[childID] = t.get(id - f.size()).clone();
                    parent1.children[childID].depth = parent1.depth + 1;
                    parent1.children[childID].parent = parent1;
                    return;
                } // Function node.
                else {
                    parent1.children[childID] = f.get(id).clone();
                    parent1.children[childID].depth = parent1.depth + 1;
                    parent1.children[childID].parent = parent1;
                }
            }

            ArrayList<Node> queue = new ArrayList();
            queue.add(parent1.children[childID]);
            Node expanded;
            int actualDepth;
            while (!queue.isEmpty()) {
                expanded = queue.get(0); // Pop from queue.
                queue.remove(0);
                actualDepth = expanded.depth + 1;
                // Create children of expanded node.
                for (int i = 0; i < expanded.arity; i++) {
                    // If depth is lower than minimal, choose from functions only.
                    if (actualDepth < SGP.MIN_INIT_DEPTH) {
                        expanded.children[i] = f.get(r.nextInt(f.size())).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                        queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                    } // Else if depth is greater than minimal and lower than maximal, choose from function and terminals.
                    else if (actualDepth < SGP.MAX_INIT_DEPTH) {
                        id = r.nextInt(f.size() + t.size());
                        // Determine if random int points to terminals or functions.
                        if (id >= f.size()) {
                            // Terminal can't be expanded, so do not add it to the queue.
                            expanded.children[i] = t.get(id - f.size()).clone();
                            expanded.children[i].depth = actualDepth;
                            expanded.children[i].parent = expanded;
                        } else {
                            expanded.children[i] = f.get(id).clone();
                            expanded.children[i].depth = actualDepth;
                            expanded.children[i].parent = expanded;
                            queue.add(expanded.children[i]); // Add it to the queue for future expansion.
                        }
                    } // Else maximal depth is reached, choose only from terminals.
                    else {
                        expanded.children[i] = t.get(r.nextInt(t.size())).clone();
                        expanded.children[i].depth = actualDepth;
                        expanded.children[i].parent = expanded;
                    }
                }
            }
        }
    }

    public int getSize() {
        int counter = 1;
        ArrayList<Node> queue1 = new ArrayList();
        queue1.add(this.rootNode);
        Node actual1 = queue1.get(0);


        // Go through tree by BFS and count nodes.
        while (true) {

            actual1 = queue1.get(0);
            queue1.remove(0);

            for (Node n : actual1.children) {
                queue1.add(n);
                counter++;
            }
            if (queue1.isEmpty()) {
                break;
            }

        }
        return counter;
    }
    @Override
    public Individual clone() {
        Individual ind = new Individual(false); // Empty individual;
        ind.rootNode = this.rootNode.clone();
        ind.fitness = this.fitness;
        ArrayList<Node> origQueue = new ArrayList(); // BFS queue.
        ArrayList<Node> newQueue = new ArrayList(); // BFS queue.
        origQueue.add(this.rootNode);
        newQueue.add(ind.rootNode);
        Node origExpanded;              // Expanded node.
        Node newExpanded;

        while (!origQueue.isEmpty()) {
            origExpanded = origQueue.get(0); // Pop from queue.
            newExpanded = newQueue.get(0); // Pop from queue.
            origQueue.remove(0);
            newQueue.remove(0);
            // Create children of expanded node.
            for (int i = 0; i < origExpanded.arity; i++) {
                newExpanded.children[i] = origExpanded.children[i].clone();
                newExpanded.children[i].depth = newExpanded.depth + 1;
                newExpanded.children[i].parent = newExpanded;
                origQueue.add(origExpanded.children[i]);
                newQueue.add(newExpanded.children[i]);
            }
        }

        
        return ind;
    }

    public String toString() {
        String r = "";
        treeString = "";
        recString(rootNode);
        /*Node n = rootNode;
        ArrayList<Node> queue = new ArrayList();
        queue.add(n);
        Node expand;
        while (!queue.isEmpty()) {
            expand = queue.get(0);
            queue.remove(0);

            r += expand.name + " ";

            for (int i = 0; i < expand.arity; i++) {
                queue.add(expand.children[i]);
            }
        }
        /**/
        
        return treeString;
    }
    
    public void recString(Node n) {
        treeString += "(";
        if(n.arity == 0) {
            treeString += n.name;
        }
        if(n.arity == 1) {
            treeString += n.name + "(";
            recString(n.children[0]);
        }
        if(n.arity == 2) {
            recString(n.children[0]);
            treeString += " "+n.name + " ";
            recString(n.children[1]);
        }
        treeString += ")";
    }

    @Override
    public int compareTo(Object o) {
        Individual n = (Individual) o;
        // Minimization.

        if (this.fitness < n.fitness) {
            return -1;
        }
        if (n.fitness < this.fitness) {
            return 1;
        }
        return 0;

    }
}
