/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package sgp;

import Nodes.*;
import java.util.ArrayList;
import java.util.Arrays;
/**
 *
 * @author Fanda
 */
public class SGP {

    public static ArrayList<Node> functions = new ArrayList();
    public static ArrayList<Node> terminals = new ArrayList();
    public static Individual bestIndividual;// = new Individual();
    public static Individual testIndividual; 
    
    public static int POP_SIZE = 1000;
    public static int GENERATIONS = 20000;
    public static double pM = 0.2; // Probability of crossing
    public static double pC = 0.9;
    public static int elites = 1;   // Number of elites
    public static int MIN_DEPTH = 1;
    public static int MAX_DEPTH = 4;
    public static Individual[] parents = new Individual[POP_SIZE];
    public static Individual[] offsprings = new Individual[POP_SIZE];
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        // Init functions and terminals for GP and store them.
        functions.add(new Add());
        functions.add(new Sub());
        functions.add(new Mul());
        functions.add(new Div());
        functions.add(new Sin());
        //functions.add(new Sqrt());
        
        terminals.add(new One());
        terminals.add(new Variable());
        
        Regression.generateControlPoints();
        
        // Init population.
        for(int ind = 0; ind < POP_SIZE; ind++) {
            parents[ind] = new Individual();
            parents[ind].execute();
        }
        // Sort initial population.
        Arrays.sort(parents);
        bestIndividual = parents[0].clone();
        
        for(int gen = 0; gen < GENERATIONS; gen++) {
            // Create new population.
            for(int pop = 0; pop < POP_SIZE; pop+=2) {
                // Selection.
                int p1 = tournamentSelection(3);
                int p2 = tournamentSelection(3);
                // Ensure that two selected individual will be different.
                while(p1 == p2) {
                    p2 = tournamentSelection(3);
                }
                // Clone selected individuals.
                offsprings[pop] = parents[p1].clone();
                offsprings[pop+1] = parents[p2].clone();
                
                // Cross them.
                if(Math.random() < pC) {
                    Individual.onePointCrossover(offsprings[pop], offsprings[pop+1]);
                }
                // Mutate them.
                if(Math.random() < pM) {
                    Individual.mutation(offsprings[pop]);
                }
                if(Math.random() < pM) {
                    Individual.mutation(offsprings[pop+1]);
                }
                // If new individuals are too large, reinit them.
                if(offsprings[pop].getSize() > (int)Math.pow(2.0, SGP.MAX_DEPTH)) {
                    offsprings[pop] = new Individual();
                }
                if(offsprings[pop+1].getSize() > (int)Math.pow(2.0, SGP.MAX_DEPTH)) {
                    offsprings[pop+1] = new Individual();
                }
                // Evaluate them.
                offsprings[pop].execute();
                offsprings[pop+1].execute();
                
            }
            // Sort new population.
            Arrays.sort(offsprings);
            // Replace old population.
            for(int i = elites; i < POP_SIZE; i++) {
                parents[i] = offsprings[i-elites].clone();
            }
            Arrays.sort(parents);
            
            // Save best individual.
            if(bestIndividual.fitness > parents[0].fitness || (bestIndividual.fitness == parents[0].fitness && parents[0].getSize() < bestIndividual.getSize())) {
                bestIndividual = parents[0].clone();
                System.out.println("Generation: " + gen + ", Fitness: " + bestIndividual.fitness + ", Size: " + bestIndividual.getSize());
                System.out.println(parents[0].toString());
            }
        }
    }
    
    /** 
     * Tournament selection
     * @param size size of the tournament
     * @return winner individual
     */
     public static int tournamentSelection(int size) {

        // Restart indexes.
        int index = (int) (Math.random() * (POP_SIZE));
        int winnerIndex = index;
        double winnerFitness = parents[index].fitness;

        // Choose N individuals and select best one.
        for (int individualIndex = 0; individualIndex < size - 1; individualIndex++) {
            index = (int) (Math.random() * (POP_SIZE));
                if (parents[index].fitness < winnerFitness) {
                    winnerFitness = parents[index].fitness;
                    winnerIndex = index;
                }

        }

        return winnerIndex;
    }
    
}
